use std::{
    collections::HashMap,
    io::{prelude::*, BufReader},
    net::TcpStream,
};

pub struct URL {
    raw: String,
    pub path: String,
    pub query: HashMap<String, String>,
}

impl URL {
    pub fn new(url: String) -> Option<Self> {
        let raw = url;
        let mut split = raw.split('?');

        // Query string parsing
        let mut query = HashMap::new();
        let path = split.next()?.to_string();
        let query_string = split.next();

        for pair in query_string.unwrap_or("").split('&') {
            let mut key_value = pair.split('=');
            let key = match key_value.next() {
                Some(key) => key.to_string(),
                None => continue,
            };
            // Skip empty keys should they appear
            if key.len() == 0 {
                continue;
            }


            let value = key_value
                .next()
                .unwrap_or("")
                .to_string()
                .replace("+", " ")
                .replace("%20", " ")
                .replace("%23", "#")
                .replace("%2C", ",");
            query.insert(key, value);
        }

        // Drop trailing slash but not for root path
        let path = if path.ends_with("/") && path.len() > 1 {
            path[..path.len() - 1].to_string()
        } else {
            path
        };

        Some(URL { path, query, raw })
    }

    /// Utility function to get a path segment by index
    ///
    /// This is a very simple function, and its implementation is fairly naïve
    ///
    /// index: Index of the path segment to get
    ///
    /// # Examples
    ///
    /// ```
    /// let url = URL::new("/path/to/resource").unwrap();
    /// assert_eq!(url.get_path_segment(1), Some("to"));
    /// ```
    pub fn get_path_segment(&self, index: usize) -> Option<&str> {
        self.path.split('/').nth(index + 1)
    }
}

pub struct Request {
    stream: TcpStream,
    pub method: String,
    pub url: URL,
}

pub enum RequestStatus {
    Ok(Request),
    // Send stream back to the caller to decide what to do with it
    // (usually close it)
    MalformedHTTP(TcpStream),
}

impl Request {
    pub fn parse_stream(stream: TcpStream) -> RequestStatus {
        let reader = BufReader::new(&stream);
        let mut lines = reader.lines();

        let request_line = match lines.next() {
            Some(Ok(line)) => line,
            _ => return RequestStatus::MalformedHTTP(stream),
        };

        let request_line_parts = request_line.split(" ").collect::<Vec<_>>();
        if request_line_parts.len() != 3 {
            return RequestStatus::MalformedHTTP(stream);
        }

        let (method, url_str) = (
            request_line_parts[0].to_string().clone(),
            request_line_parts[1].to_string().clone(),
        );
        let url = match URL::new(url_str) {
            Some(path) => path,
            None => return RequestStatus::MalformedHTTP(stream),
        };

        RequestStatus::Ok(Request {
            stream,
            method,
            url,
        })
    }

    pub fn mut_stream(self) -> TcpStream {
        self.stream
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_url_new() {
        let url = URL::new("/".to_string()).unwrap();
        assert_eq!(url.path, "/");
        assert_eq!(url.raw, "/");

        let url = URL::new("/path/to/resource".to_string()).unwrap();
        assert_eq!(url.path, "/path/to/resource");
        assert_eq!(url.query.len(), 0);

        let url = URL::new("/path/to/resource?query=string".to_string()).unwrap();
        assert_eq!(url.path, "/path/to/resource");
        assert_eq!(url.query.len(), 1);
        assert_eq!(url.query.get("query").unwrap(), "string");

        let url = URL::new("/path/to/resource?query=string&another=one".to_string()).unwrap();
        assert_eq!(url.path, "/path/to/resource");
        assert_eq!(url.query.len(), 2);
        assert_eq!(url.query.get("query").unwrap(), "string");
        assert_eq!(url.query.get("another").unwrap(), "one");

        let url = URL::new("/path/to/resource?query=string&another=one&third=3".to_string()).unwrap();
        assert_eq!(url.path, "/path/to/resource");
        assert_eq!(url.query.len(), 3);
        assert_eq!(url.query.get("query").unwrap(), "string");
        assert_eq!(url.query.get("another").unwrap(), "one");
        assert_eq!(url.query.get("third").unwrap(), "3");
    }
}
